#include "suite/suite.h"

#include "source/forstio/proto_kel.h"

#include <iostream>

namespace {
namespace schema {
	using namespace saw::schema;
}
using TestSize = schema::UInt32;

using TestTuple = schema::Tuple<schema::UInt32, schema::UInt16>;

using TestStruct = schema::Struct<
	schema::NamedMember<schema::UInt32, "test_uint">,
	schema::NamedMember<schema::String, "test_string">,
	schema::NamedMember<schema::String, "test_name">
>;

using TestUnion = schema::Union<
	schema::NamedMember<schema::UInt32, "test_uint">,
	schema::NamedMember<schema::String, "test_string">
>;

SAW_TEST("Primitive Encoding"){
	using namespace saw;
	uint32_t value = 5;

	auto root = new_heap_message_root<TestSize>();
	auto builder = root.build();

	builder.set(value);

	ring_buffer temp_buffer;
	proto_kel_codec codec;

	error err = codec.encode<TestSize>(root.read(), temp_buffer);

	SAW_EXPECT(!err.failed(), err.message());
	SAW_EXPECT(temp_buffer.read_composite_length() == (sizeof(value)+sizeof(proto_kel_codec::PacketLengthT)), "Bad Size: " + std::to_string(temp_buffer.read_composite_length()));
	constexpr size_t pkt_shift = sizeof(proto_kel_codec::PacketLengthT);
	SAW_EXPECT(temp_buffer[pkt_shift] == 5 && temp_buffer[pkt_shift+1] == 0 && temp_buffer[pkt_shift+2] == 0 && temp_buffer[pkt_shift+3] == 0, "Wrong encoded values");
}

SAW_TEST("List Encoding"){
	using namespace saw;

	auto root = new_heap_message_root<TestTuple>();
	auto builder = root.build();

	auto first = builder.init<0>();
	first.set(2135231);
	auto second = builder.init<1>();
	second.set(43871);

	ring_buffer buffer;
	proto_kel_codec codec;

	error err = codec.encode<TestTuple>(root.read(), buffer);

	SAW_EXPECT(!err.failed(), err.message());
	SAW_EXPECT(buffer.read_composite_length() == 14, "Bad Size: " + std::to_string(buffer.read_composite_length()));
	SAW_EXPECT("06 00 00 00\n00 00 00 00\nbf 94 20 00\n5f ab" == buffer.to_hex(), "Not equal encoding\n"+buffer.to_hex());
}

SAW_TEST("Struct Encoding"){
	using namespace saw;

	auto root = new_heap_message_root<TestStruct>();
	auto builder = root.build();

	auto test_uint = builder.init<"test_uint">();
	test_uint.set(23);

	std::string test_string = "foo";
	auto string = builder.init<"test_string">();
	string.set(test_string);
	
	auto string_name = builder.init<"test_name">();
	string_name.set("test_name");

	ring_buffer buffer;
	proto_kel_codec codec;

	error err = codec.encode<TestStruct>(builder.as_reader(), buffer);

	SAW_EXPECT(!err.failed(), err.message());
	SAW_EXPECT(buffer.read_composite_length() == 40, "Bad Size: " + std::to_string(buffer.read_composite_length()));
	SAW_EXPECT("20 00 00 00\n00 00 00 00\n17 00 00 00\n03 00 00 00\n00 00 00 00\n66 6f 6f 09\n00 00 00 00\n00 00 00 74\n65 73 74 5f\n6e 61 6d 65"
		== buffer.to_hex(), "Not equal encoding:\n"+buffer.to_hex());
}

SAW_TEST("Union Encoding"){
	using namespace saw;
	{
		auto root = new_heap_message_root<TestUnion>();
		auto builder = root.build();

		auto test_uint = builder.init<"test_uint">();
		test_uint.set(23);

		ring_buffer buffer;
		proto_kel_codec codec;

		error err = codec.encode<TestUnion>(builder.as_reader(), buffer);

		SAW_EXPECT(!err.failed(), err.message());
		SAW_EXPECT(buffer.read_composite_length() == 16, "Bad Size: " + std::to_string(buffer.read_composite_length()));
		SAW_EXPECT("08 00 00 00\n00 00 00 00\n00 00 00 00\n17 00 00 00"
			== buffer.to_hex(), "Not equal encoding:\n"+buffer.to_hex());
	}
	{
		auto root = new_heap_message_root<TestUnion>();
		auto builder = root.build();

		auto test_string = builder.init<"test_string">();
		test_string.set("foo");

		ring_buffer buffer;
		proto_kel_codec codec;

		error err = codec.encode<TestUnion>(builder.as_reader(), buffer);

		SAW_EXPECT(!err.failed(), err.message());
		SAW_EXPECT(buffer.read_composite_length() == 23, "Bad Size: " + std::to_string(buffer.read_composite_length()));
		SAW_EXPECT("0f 00 00 00\n00 00 00 00\n01 00 00 00\n03 00 00 00\n00 00 00 00\n66 6f 6f"
			== buffer.to_hex(), "Not equal encoding:\n"+buffer.to_hex());
	}
}

SAW_TEST("Tuple Decoding"){
	using namespace saw;
	const uint8_t buffer_raw[] = {0x06, 0, 0, 0, 0, 0, 0, 0, 0xbf, 0x94, 0x20, 0x00, 0x5f, 0xab};

	ring_buffer buffer;
	buffer.push(*buffer_raw, sizeof(buffer_raw));

	proto_kel_codec codec;

	auto root = new_heap_message_root<TestTuple>();
	auto builder = root.build();

	error err = codec.decode<TestTuple>(builder, buffer);
	SAW_EXPECT(!err.failed(), err.message());
	
	auto reader = builder.as_reader();

	auto first = reader.get<0>();
	auto second = reader.get<1>();

	SAW_EXPECT(first.get() == 2135231 && second.get() == 43871, "Values not correctly decoded");
}

SAW_TEST("Struct Decoding"){
	using namespace saw;
	const uint8_t buffer_raw[] = {0x20,0,0,0,0,0,0,0,0x17,0,0,0,0x03,0,0,0,0,0,0,0,0x66,0x6f,0x6f,0x09,0,0,0,0,0,0,0,0x74,0x65,0x73,0x74,0x5f,0x6e,0x61,0x6d,0x65};

	ring_buffer buffer;
	buffer.push(*buffer_raw, sizeof(buffer_raw));

	proto_kel_codec codec;

	auto root = new_heap_message_root<TestStruct>();
	auto builder = root.build();

	error err = codec.decode<TestStruct>(builder, buffer);
	auto reader = builder.as_reader();

	auto foo_string = reader.get<"test_string">();
	auto test_uint = reader.get<"test_uint">();
	auto test_name = reader.get<"test_name">();

	SAW_EXPECT(!err.failed(), err.message());
	SAW_EXPECT(foo_string.get() == "foo" && test_uint.get() == 23 && test_name.get() == "test_name", "Values not correctly decoded");
}

SAW_TEST("Union Decoding"){
	using namespace saw;
	const uint8_t buffer_raw[] = {0x0f,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x01,0x00,0x00,0x00,0x03,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x66,0x6f,0x6f};

	ring_buffer buffer;
	buffer.push(*buffer_raw, sizeof(buffer_raw));

	proto_kel_codec codec;

	auto root = new_heap_message_root<TestUnion>();
	auto builder = root.build();
	auto reader = builder.as_reader();

	error err = codec.decode<TestUnion>(builder, buffer);

	SAW_EXPECT(!err.failed(), err.message());
	SAW_EXPECT(reader.has_alternative<"test_string">(), "Wrong union value");
	auto str_rd = reader.get<"test_string">();
	SAW_EXPECT(str_rd.get() == "foo", "Wrong value: " + std::string{str_rd.get()});
}

using TestArrayStruct = schema::Array<
	TestStruct
>;

SAW_TEST("Array Encoding"){
	using namespace saw;
	
	proto_kel_codec codec;
	auto root = new_heap_message_root<TestArrayStruct>();
	auto builder = root.build(2);

	auto one = builder.init(0);
	auto two = builder.init(1);

	one.init<"test_uint">().set(4);
	one.init<"test_string">().set("foo");
	one.init<"test_name">().set("Fedor");
	
	two.init<"test_uint">().set(9);
	two.init<"test_string">().set("bar");
	two.init<"test_name">().set("Bravo");

	ring_buffer buffer;

	error err = codec.encode<TestArrayStruct>(root.read(), buffer);

	SAW_EXPECT(!err.failed(), "Error occured");
}
}
